import os
import subprocess
import sys
from pathlib import Path

from . import support
from .. import api
from .click import api as click


default_help = """
\b 
Available Virtual Environments:
    {envs}

""".format(envs="\n    ".join(f'{name:<12}' for name in sorted(api.find_existing_venv_names())))


@click.command(help=default_help, context_settings={'ignore_unknown_options': True, 'allow_interspersed_args': False})
@click.option('-c', '--copy', is_flag=True, help='Do not create symlinks for python')
@click.option('-C', '--create-only', is_flag=True, help='Only create venv, do not enter')
@click.option('-d', '--dry-run', is_flag=True, help='Do not make changes to the system')
@click.option('-e', '--ephemeral', is_flag=True, help='Create and remove')
@click.option('-i', '--interactive', is_flag=True, help='Run interactively')
@click.option('-l', '--ls', is_flag=True, help='Show available virtual environments')
@click.option('--no-pip', is_flag=True, help='Do not include pip')
@click.option('-o', '--overwrite', is_flag=True, help='Recreate venv')
@click.option('--path', metavar='PATH', help='Path to virtual environment')
@click.option('-p', '--python', metavar='VERSION', help='Python version to use')
@click.option('-r', '--remove', is_flag=True, help='Remove virtual enironment')
@click.option('-u', '--upgrade', is_flag=True, help='Upgrades to latest python version')
@click.option('-v', '--verbose', count=True, help='More output')
@click.option('-V', '--version', is_flag=True, help='Show version and exit')
@click.option('--shell-completion', is_flag=True, help='Show shell completion code')
@click.argument('name', metavar='VENV_NAME', type=click.OptionalChoice(api.find_existing_venv_names()), nargs=1, required=False)
@click.argument('command', required=False, nargs=-1)
@click.pass_context
def vsh(ctx, copy, create_only, dry_run, ephemeral, interactive, shell_completion, ls, no_pip, overwrite, path, python, remove, upgrade, verbose, version, name, command):
    if shell_completion:
        subprocess.run('_VSH_COMPLETE=source vsh', shell=True)
        sys.exit(0)

    verbose = verbose or 0
    return_code = 0
    if version:
        api.show_version()
        sys.exit(0)

    if ls:
        api.show_envs()
        sys.exit(0)

    if path and name:
        # favor path over name
        command = [name] + list(command)
        name = Path(path).name

    if not (path or name):
        click.echo(vsh.get_help(ctx))
        click.echo('\nERROR: A name or path must be provided.')
        sys.tracebacklimit = 0
        sys.exit(1)

    if not path:
        home = os.getenv('HOME')
        workon_home = os.getenv('WORKON_HOME') or os.path.join(home, '.virtualenvs')
        path = os.path.join(workon_home, name)

    # Determine if an environment already exists
    exists = api.validate_environment(path)

    if not command and not remove:
        command = os.getenv('SHELL')

    if exists and upgrade:
        # TODO: Add this
        #       Use pip freeze to capture everything installed currently
        #       Modify the python used
        #       Use pip and freeze list to rebuild virtual env with new python
        pass

    elif not exists and not remove:
        api.create(path, include_pip=not no_pip, overwrite=overwrite, symlinks=not copy, python=python, verbose=verbose)
        if ephemeral:
            remove = True

    if command and not create_only:
        return_code = api.enter(path, command, verbose=max(verbose - 1, 0))

    if ephemeral and not remove:
        quoted_name = '"{name}"'.format(name=click.style(name, fg="yellow"))
        rm_command = click.style(f'vsh -r {name}', fg='blue')
        support.echo(f'Virtual environment {quoted_name} existed previously.  Aborting removal.\nTo remove, run:\n\n    {rm_command}\n')

    if remove:
        api.remove(path, verbose=verbose, interactive=interactive, dry_run=dry_run)

    sys.tracebacklimit = 0
    sys.exit(return_code)



from pkg_resources import parse_version

__all__ = (
    '__name__', '__description__', '__version__', '__author__',
    '__author_email__', '__maintainer__', '__maintainer_email__',
    '__copyright_years__', '__license__', '__url__', '__version_info__',
    '__classifiers__', '__keywords__', 'package_metadata',
    )

# ----------------------------------------------------------------------
# Package Metadata (Update manually)
# ----------------------------------------------------------------------
__name__ = 'vsh'
__description__ = 'A virtual environment shell'
__ver__ = '0.6.1'  # Must follow pep440

__author__ = 'Brian Bruggeman'
__author_email__ = 'brian.m.bruggeman@gmail.com'

__maintainer__ = 'Brian Bruggeman'
__maintainer_email__ = 'brian.m.bruggeman@gmail.com'

__copyright_years__ = '2018'
__license__ = 'MIT'
__url__ = 'https://github.com/brianbruggeman/vsh'

__keywords__ = ['vsh', 'virtual environment', 'venv']


# See: https://pypi.python.org/pypi?%3Aaction=list_classifiers
__classifiers__ = [
    'Programming Language :: Python',
    'Natural Language :: English',
    'Development Status :: 3 - Alpha',
    'Natural Language :: English',
    'Operating System :: POSIX',
    'Programming Language :: Python :: 3.6',
    ]


# ----------------------------------------------------------------------
# Programmatic values (Do not update)
# ----------------------------------------------------------------------
__version__ = parse_version(__ver__).public
__version_info__ = tuple(
    int(ver_i)
    for ver_i in parse_version(__version__).base_version.split('.')
    if ver_i.isdigit()
    )


# Package metadata is used in setup
package_metadata = {
    k.strip('_'): v
    for k, v in locals().items()
    if k.startswith('__')
    # Clean items that don't mix with setup
    if k.strip('_') not in ['all', 'copyright_years', 'ver', 'version_info']
    }



#!/usr/bin/env python3
import errno
import os
import select
import shlex
import sys
from subprocess import PIPE, Popen

if sys.version.startswith('3'):
    basestring = str


def run(command, env=None, buffer_size=None):
    """Run the given command.

    Args:
        command(str or list): command
        env(dict): Environment variables to use
        cwd(str): Path to execute this command

    Returns:
        int: return code for process
    """
    command = shlex.split(command) if not isinstance(command, list) else command
    env = env or os.environ.copy()
    process = Popen(command, env=env, stdin=PIPE, stdout=PIPE, stderr=PIPE)
    pipes = [process.stderr, process.stdout, process.stdin]
    filenos = [_.fileno() for _ in pipes]
    should_run = True
    fd_mapping = {
        process.stdout: sys.stdout,
        process.stderr: sys.stderr,
        # process.stdin: sys.stdin
        }
    while should_run:
        try:
            selectables = select.select(filenos, filenos, filenos)
        except select.error as e:
            if e.args[0] == errno.EINTR:
                continue
        if not any(selectables):
            continue
        for selected_list in selectables:
            for fileno in selected_list:
                for fd in pipes:
                    if fd.fileno() != fileno:
                        continue
                    rfd = fd_mapping.get(fd)
                    if rfd is None:
                        continue
                    for line in fd:
                        yield rfd, line.decode('utf-8')

        if process.poll() is not None:
            should_run = False
    yield None, process.returncode


def execute_hooks():
    hook = os.path.abspath(sys.argv[0])
    folder = hook + '.d'
    if os.path.exists(folder) and os.path.isdir(folder):
        for root, folders, files in os.walk(folder):
            for filename in files:
                filepath = os.path.join(root, filename)
                if os.access(filepath, os.X_OK):
                    for fd, line in run(filepath):
                        if fd is None and isinstance(line, int):
                            if line != 0:
                                return line
                            continue
                        fd.write(line)


if __name__ == '__main__':
    import sys
    sys.exit(execute_hooks() or 0)



# -*- coding: utf-8 -*-
"""
    click
    ~~~~~

    Click is a simple Python module that wraps the stdlib's optparse to make
    writing command line scripts fun.  Unlike other modules, it's based around
    a simple API that does not come with too much magic and is composable.

    In case optparse ever gets removed from the stdlib, it will be shipped by
    this module.

    :copyright: (c) 2014 by Armin Ronacher.
    :license: BSD, see LICENSE for more details.
"""
# Core classes
from .core import (Argument, BaseCommand, Command, CommandCollection, Context,
                   Group, MultiCommand, Option, Parameter)
# Decorators
from .decorators import (argument, command, confirmation_option, group,
                         help_option, make_pass_decorator, option,
                         pass_context, pass_obj, password_option,
                         version_option)
# Exceptions
from .exceptions import (Abort, BadArgumentUsage, BadOptionUsage, BadParameter,
                         ClickException, FileError, MissingParameter,
                         NoSuchOption, UsageError)
# Formatting
from .formatting import HelpFormatter, wrap_text

# Globals
from .globals import get_current_context

# Parsing
from .parser import OptionParser

# Terminal functions
from .termui import (clear, confirm, echo_via_pager, edit, get_terminal_size,
                     getchar, launch, pause, progressbar, prompt, secho, style,
                     unstyle)

# Types
from .types import (BOOL, FLOAT, INT, STRING, UNPROCESSED, UUID, Choice, OptionalChoice, File,
                    IntRange, ParamType, Path, Tuple)

# Utilities
from .utils import (echo, format_filename, get_app_dir, get_binary_stream,
                    get_os_args, get_text_stream, open_file)


__all__ = [
    # Core classes
    'Context', 'BaseCommand', 'Command', 'MultiCommand', 'Group',
    'CommandCollection', 'Parameter', 'Option', 'Argument',

    # Globals
    'get_current_context',

    # Decorators
    'pass_context', 'pass_obj', 'make_pass_decorator', 'command', 'group',
    'argument', 'option', 'confirmation_option', 'password_option',
    'version_option', 'help_option',

    # Types
    'ParamType', 'File', 'Path', 'Choice', 'IntRange', 'Tuple', 'STRING',
    'INT', 'FLOAT', 'BOOL', 'UUID', 'UNPROCESSED',

    # Utilities
    'echo', 'get_binary_stream', 'get_text_stream', 'open_file',
    'format_filename', 'get_app_dir', 'get_os_args',

    # Terminal functions
    'prompt', 'confirm', 'get_terminal_size', 'echo_via_pager',
    'progressbar', 'clear', 'style', 'unstyle', 'secho', 'edit', 'launch',
    'getchar', 'pause',

    # Exceptions
    'ClickException', 'UsageError', 'BadParameter', 'FileError',
    'Abort', 'NoSuchOption', 'BadOptionUsage', 'BadArgumentUsage',
    'MissingParameter',

    # Formatting
    'HelpFormatter', 'wrap_text',

    # Parsing
    'OptionParser',
    ]


# Controls if click should emit the warning about the use of unicode
# literals.
disable_unicode_literals_warning = False


__version__ = '6.7'



import itertools
import os
import re
import shlex
import shutil
import subprocess
import sys
import types
import venv
from pathlib import Path

from .__metadata__ import package_metadata
from .cli import support
from .cli.click import api as click
from .errors import InterpreterNotFound, InvalidEnvironmentError, PathNotFoundError

__all__ = ('create', 'enter', 'remove', 'show_envs', 'show_version')


class VenvBuilder(venv.EnvBuilder):

    def create(self, env_dir, executable=None):
        """
        Create a virtual environment in a directory.

        Args:
            env_dir (str): The target directory to create an environment in.
            executable (str, optional): path to python interpreter executable [default: sys.executable]
        """
        env_dir = os.path.abspath(env_dir)
        context = self.ensure_directories(env_dir=env_dir, executable=executable)
        # See issue 24875. We need system_site_packages to be False
        # until after pip is installed.
        true_system_site_packages = self.system_site_packages
        self.system_site_packages = False
        self.create_configuration(context)
        self.setup_python(context)
        if self.with_pip:
            self._setup_pip(context)
        if not self.upgrade:
            self.setup_scripts(context)
            self.post_setup(context)
        if true_system_site_packages:
            # We had set it to False before, now
            # restore it and rewrite the configuration
            self.system_site_packages = True
        self.create_configuration(context)

    def ensure_directories(self, env_dir, executable=None):
        """
        Create the directories for the environment.
        Returns a context object which holds paths in the environment,
        for use by subsequent logic.

        Args:
            env_dir (str): path to environment
            executable (str, optional): path to python interpreter executable [default: sys.executable]

        Returns:
            types.SimpleNamespace: context
        """
        env_dir = Path(env_dir)

        def create_if_needed(d):
            d = Path(d)
            if d and not d.exists():
                d.mkdir(parents=True, exist_ok=True)
            if not d or not d.is_dir():
                raise ValueError(f'Unable to create directory {d!r}')

        executable = executable or sys.executable
        if env_dir.exists() and self.clear:
            self.clear_directory(env_dir)
        context = types.SimpleNamespace()
        context.env_dir = str(env_dir)
        context.env_name = env_dir.stem
        prompt = self.prompt if self.prompt is not None else context.env_name
        context.prompt = f'({prompt}) '
        create_if_needed(env_dir)
        dirname, exename = os.path.split(os.path.abspath(executable))
        context.executable = executable
        context.python_dir = dirname
        context.python_exe = exename
        if sys.platform == 'win32':
            binname = 'Scripts'
            incpath = 'Include'
            libpath = env_dir / 'Lib' / 'site-packages'
        else:
            binname = 'bin'
            incpath = 'include'
            libpath = env_dir / 'lib' / exename / 'site-packages'
        path = env_dir / incpath
        context.inc_path = str(path)
        create_if_needed(path)
        create_if_needed(libpath)
        # Issue 21197: create lib64 as a symlink to lib on 64-bit non-OS X POSIX
        if (sys.maxsize > 2**32) and (os.name == 'posix') and (sys.platform != 'darwin'):
            link_path = os.path.join(env_dir, 'lib64')
            if not os.path.exists(link_path):   # Issue #21643
                os.symlink('lib', link_path)
        binpath = env_dir / binname
        context.bin_path = str(binpath)
        context.bin_name = binname
        context.env_exe = str(binpath / exename)
        create_if_needed(binpath)
        return context

    def _setup_pip(self, context):
        """Installs or upgrades pip in a virtual environment"""
        # We run ensurepip in isolated mode to avoid side effects from
        # environment vars, the current directory and anything else
        # intended for the global Python environment
        # Originally -Im, but -Esm works on both python2 and python3
        cmd = [context.env_exe, '-Esm', 'ensurepip', '--upgrade',
                                                     '--default-pip']
        subprocess.check_output(cmd, stderr=subprocess.STDOUT)


def build_vsh_config_file(venv_path, startup_path=None):
    """Sets a default configuration.

    When the virtual environment is started, this particular file will
    set the current working folder to startup_path.

    Args:
        venv_path (str|Path): path to virtual environment
        startup_path (str|Path): path to startup folder
    """
    default_startup_path = Path('.')
    startup_path = Path(startup_path or default_startup_path)
    vsh_venv_config_path = venv_path / '.vshrc'
    if vsh_venv_config_path.parent.exists() and not vsh_venv_config_path.exists():
        if startup_path.exists() and startup_path.is_dir():
            with vsh_venv_config_path.open('w') as config:
                config.write(f'cd {startup_path.absolute()}\n')
            support.echo(f'Set default path to: {click.style(str(startup_path), fg="blue")}')
            support.echo(f'To edit, update: {click.style(str(vsh_venv_config_path), fg="yellow")}')


def create(path, site_packages=None, overwrite=None, symlinks=None, upgrade=None, include_pip=None, prompt=None, python=None, verbose=None, interactive=None, dry_run=None):
    """Creates a virtual environment

    Notes: Wraps venv

    Args:
        path (str): path to virtual environment

        site_packages (bool, optional): use system packages within environment [default: False]
        overwrite (bool, optional): replace target folder [default: False]
        symlinks (bool, optional): create symbolic link to Python executable [default: True]
        upgrade (bool, optional): Upgrades existing environment with new Python executable [default: False]
        include_pip (bool, optional): Includes pip within virtualenv [default: True]
        prompt (str, optional): Modifies prompt
        python (str, optional): Version of python, python executable or path to python

        verbose (int, optional): more output [default: 0]
        interactive (bool, optional): ask before updating system [default: False]
        dry_run (bool, optional): do not update system

    Returns:
        str: path to venv
    """
    verbose = max(int(verbose or 0), 0)
    path = os.path.expanduser(path) if path.startswith('~') else os.path.abspath(path)
    name = os.path.basename(path)
    builder = _get_builder(path=path, site_packages=site_packages, overwrite=overwrite, symlinks=symlinks, upgrade=upgrade, include_pip=include_pip, prompt=prompt)
    prompt = f'Create virtual environment "{name}" under: {path}?'
    run_command = click.confirm(prompt) if interactive else True
    if run_command:
        if not dry_run:
            executable = _get_interpreter(python)
            if not executable:
                raise InterpreterNotFound(version=python)
            builder.create(env_dir=path, executable=executable)
        support.echo('Created virtual environment "' + click.style(name, fg='yellow') + " under: " + click.style(path, fg='green'), verbose=verbose)
    return path


def enter(path, command=None, verbose=None):
    """Enters a virtual environment

    Args:
        path (str): path to virtual environment
        command (tuple|list|str, optional): command to run in virtual env [default: shell]
        verbose (int, optional): Adds more information to stdout
    """
    verbose = max(int(verbose or 0), 0)
    path = os.path.expanduser(path) if path.startswith('~') else os.path.abspath(path)
    shell = os.getenv("SHELL")
    command = command or shell
    env = _update_environment(path)
    venv_name = click.style(Path(path).name, fg='green')

    # Setup the environment scripts
    vshell_config_commands = '; '.join(f'source {filepath}' for filepath in find_vsh_config_files(path))
    if not isinstance(command, str):
        command = " ".join(command)
    if vshell_config_commands:
        command = f'{vshell_config_commands}; {command}'
    cmd_display = click.style(command, fg='green')
    if Path(shell).name in ['bash', 'zsh']:
        command = f'{shell} -i -c \"{command}\"'
        cmd_display = f'{shell} -i -c \"{cmd_display}\"'

    support.echo(click.style(f'Running command in "', fg='blue') + venv_name + click.style(f'": ', fg='blue') + cmd_display, verbose=max(verbose - 1, 0))

    # Activate and run
    return_code = subprocess.run(command, shell=True, env=env, universal_newlines=True)
    rc_color = 'green' if return_code == 0 else 'red'
    rc = click.style(str(return_code), fg=rc_color)
    support.echo(click.style('Command return code: ', fg='blue') + rc, verbose=verbose)
    return return_code


def find_vsh_config_files(venv_path=None):
    cmds = [
        'git rev-parse --show-toplevel',
        'hg root',
        ]
    top_of_current_repo_path = None
    for cmd in cmds:
        cmd = shlex.split(cmd)
        try:
            top_of_current_repo_path = Path(subprocess.run(cmd, stderr=subprocess.PIPE, stdout=subprocess.PIPE).stdout.decode('utf-8').strip()) or None
            if top_of_current_repo_path:
                break
        except Exception:
            pass
    paths = [
        Path('/usr/local/etc/vsh'),
        Path(os.getenv('HOME')),
        Path('.'),
        Path(venv_path),
        top_of_current_repo_path,
        ]
    memoized_paths = set()
    for p in paths:
        if p is None:
            continue
        config_path = p / '.vshrc'
        if not config_path.exists() and str(p) == venv_path:
            startup_path = Path(top_of_current_repo_path or '.')
            build_vsh_config_file(p, startup_path=startup_path)
        if p.exists() and config_path.exists():
            if config_path.is_file():
                if config_path in memoized_paths:
                    continue
                memoized_paths.add(config_path)
                yield config_path
            elif config_path.is_dir():
                for root, folders, files in os.walk(str(config_path)):
                    root = Path(root)
                    for filename in files:
                        filepath = (root / filename).absolute()
                        if filepath in memoized_paths:
                            continue
                        memoized_paths.add(filepath)
                        yield filepath


def find_environment_folders(path=None):
    path = path or os.getenv('WORKON_HOME') or os.path.join(os.getenv('HOME'), '.virtualenvs')
    for root, directories, files in os.walk(path):
        found = []
        for index, name in enumerate(directories):
            directory = os.path.join(root, name)
            if not validate_environment(directory):
                continue
            yield name, directory
            found.append(name)
        # This makes the search "fast" by skipping out on folders
        #  that do not need to be searched because they have already
        #  been identified as valid environments
        directories[:] = [d for d in directories if d not in found]


def find_existing_venv_names(venvs_home=None):
    home = os.getenv('HOME')
    workon_home = os.getenv('WORKON_HOME') or Path(f'{home}/.virtualenvs')
    venvs_home = venvs_home or Path(workon_home) if workon_home and Path(workon_home).exists() else None

    if venvs_home and venvs_home.exists():
        standard_path = ['include', 'lib', 'bin']
        for path in os.scandir(venvs_home):
            if Path(path).is_dir():
                if Path(path).stem not in standard_path:
                    yield Path(path).stem


def remove(path, verbose=None, interactive=None, dry_run=None, check=None):
    """Remove a virtual environment

    Args:
        path (str): path to virtual environment
        verbose (int, optional): more output [default: 0]
        interactive (bool, optional): ask before updating system [default: False]
        dry_run (bool, optional): do not update system
        check (bool, optional): Raises PathNotFoundError if True and path isn't found [default: False]

    Raises:
        PathNotFoundError:  when check is True and path is not found

    Returns:
        str: folder path removed
    """
    verbose = max(int(verbose or 0), 0)
    check = False if check is None else check
    path = os.path.expanduser(path) if path.startswith('~') else os.path.abspath(path)
    if not validate_environment(path) and check is True:
        raise InvalidEnvironmentError(path=path)
    prompt = f'Remove {path}?'
    run_command = click.confirm(prompt) == 'y' if interactive else True
    if run_command and not dry_run:
        if os.path.exists(path):
            shutil.rmtree(path)
        elif check is True:
            raise PathNotFoundError(path=path)
    support.echo(click.style('Removed: ', fg='blue') + click.style(path, fg='green'), verbose=(max(verbose - 1, 0) and path))
    return path


def show_envs(path=None):
    path = path or os.getenv('WORKON_HOME')
    for name, path in find_environment_folders(path=path):
        print(f'Found {click.style(name, fg="yellow")} under: {click.style(path, fg="yellow")}')


def show_version():
    support.echo(f"{package_metadata['name']} {package_metadata['version']}")


def validate_environment(path, check=None):
    """Validates if path is a virtual environment

    Args:
        path (str): path to virtual environment
        check (bool, optional): Raise an error if path isn't valid

    Raises:
        InvalidEnvironmentError: when environment is not valid

    Returns:
        bool: True if valid virtual environment path
    """
    path = Path(path)
    valid = None
    win32 = sys.platform == 'win32'
    # Expected structure
    structure = {
        'bin': 'Scripts' if win32 else 'bin',
        'include': 'Include' if win32 else 'include',
        'lib': os.path.join('Lib', 'site-packages') if win32 else os.path.join('lib', '*', 'site-packages'),
        }
    paths = {}
    for identifier, expected_path in structure.items():
        for p in path.glob(expected_path):
            # There should only be one path that matches the glob
            paths[identifier] = p
            break
    for identifier in structure:
        if identifier not in paths:
            valid = False
            if check:
                raise InvalidEnvironmentError(f'Could not find {structure[identifier]} under {path}.')

    if valid is not False and win32:
        # TODO: Add more validation for windows environments
        valid = valid is not False and True
    elif valid is not False:
        # check for activation scripts
        activation_scripts = list(paths['bin'].glob('activate.*'))
        valid = valid is not False and len(activation_scripts) > 0
        if check and valid is False:
            raise InvalidEnvironmentError(f'Could not find activation scripts under {path}.')

        # check for python binaries
        python_name = paths['lib'].parent.name
        python_ver_data = re.search('(?P<interpreter>python|pypy)\.?(?P<major>\d+)(\.?(?P<minor>\d+))', python_name)
        if python_ver_data:
            python_ver_data = python_ver_data.groupdict()
            python_executable = paths['bin'].joinpath('python')
            python_ver_executable = paths['bin'].joinpath(python_name)
            if python_executable.exists():
                valid = valid is not False and True
            if check and valid is False:
                raise InvalidEnvironmentError(f'Could not find python executable under {path}.')
            if python_ver_executable.exists():
                valid = valid is not False and True
            if check and valid is False:
                raise InvalidEnvironmentError(f'Could not find {python_name} executable under {path}.')

    return valid


# ----------------------------------------------------------------------
# Support
# ----------------------------------------------------------------------
def _escape_zero_length_codes(prompt=None):
    # This is necessary because bash does something funky with PS1 and
    #  doesn't correctly calculate the length of the command-line.  When
    #  under a TMUX session, the command-line itself ends up getting
    #  corrupted when looking at history
    start = r'(\\e|\\xb1)\['
    directive = r'[^m]*'
    end = r'm'
    pattern = f'({start}{directive}{end})'
    replacement = r'\[\1\]'

    eng = re.compile(pattern)
    prompt = prompt or os.getenv('PS1') or os.getenv('PROMPT')
    if prompt:
        prompt = eng.sub(replacement, prompt)
    else:
        prompt = r'\[\e[34m\]\w\[\e[0m\] \[\e[33m\]\$\[\e[0m\]'
    return prompt


def _get_builder(path, site_packages=None, overwrite=None, symlinks=None, upgrade=None, include_pip=None, prompt=None):
    path = os.path.expanduser(path) if path.startswith('~') else os.path.abspath(path)
    name = os.path.basename(path)
    builder = VenvBuilder(
        system_site_packages=False if site_packages is None else site_packages,
        clear=False if overwrite is None else overwrite,
        symlinks=True if symlinks is None else symlinks,
        upgrade=False if upgrade is None else upgrade,
        with_pip=True if include_pip is None else include_pip,
        prompt=f'({name})' if prompt is None else prompt,
        )
    return builder


def _get_interpreter(python=None):
    """Returns the interpreter given the string"""
    if python is None:
        return sys.executable

    # Maybe the path is already supplied
    if Path(python).exists():
        return python

    # Guess path
    paths = [Path(path) for path in os.getenv('PATH').split(':')]

    # Assume that python is a version if it doesn't start with p
    python = f'python{python}' if not python.startswith('p') else python
    interpreters = [python]

    # Build potential interpreter paths
    interpreter_paths = [p / i for p in paths for i in interpreters]
    for path in interpreter_paths:
        if path.absolute().exists():
            # return the first one found
            return path


def _update_environment(path):
    """Updates environment similar to activate from venv"""
    path = os.path.expanduser(path) if path.startswith('~') else os.path.abspath(path)
    name = os.path.basename(path)

    env = {k: v for k, v in os.environ.items()}
    env[package_metadata['name'].upper()] = name

    env['VIRTUAL_ENV'] = path
    env['PATH'] = ':'.join([os.path.join(env.get('VIRTUAL_ENV'), 'bin')] + env['PATH'].split(':'))

    shell = Path(env.get('SHELL') or '/bin/sh').name
    disable_prompt = env.get('VIRTUAL_ENV_DISABLE_PROMPT') or None
    shell_prompt_mapping = {
        'bash': 'PS1',
        'sh': 'PS1',
        'zsh': 'PROMPT'
        }
    shell_prompt_var = shell_prompt_mapping.get(shell)
    prompt = env.get(shell_prompt_var) or click.style("\w", fg="blue") + "\$ "
    prompt = _escape_zero_length_codes(prompt) if shell in ['bash', 'sh'] else prompt
    if shell_prompt_var and not disable_prompt:
        env[shell_prompt_var] = prompt

    return env



class BaseError(Exception):
    """Fancy error handler"""

    def __init__(self, message=None, **kwds):
        if message is not None:
            message = message.format(**kwds) if kwds else message
        else:
            message = self.__doc__.format(**kwds) if kwds else self.__doc__
        self.__msg__ = message
        Exception.__init__(self, message)

    def __repr__(self):
        cname = type(self).__name__
        message = self.__msg__
        return f'<{cname} {message}>'

    def __str__(self):
        return self.__msg__


class InterpreterNotFound(BaseError):
    """ERROR: Could not find interpreter for: {version}"""


class InvalidEnvironmentError(BaseError):
    """ERROR: Path is not a valid environment: {path}"""


class PathNotFoundError(BaseError):
    """ERROR: Could not find path: {path}"""



#!/usr/bin/env python
"""
Setup.py shouldn't be updated during normal development.

Update:
   entrypoints.txt  <-- add new command-line interfaces here
   requirements.txt  <-- required for package to run
   requirements/*.txt  <-- as appropriate
"""
import datetime
import itertools
import os
import shutil
import sys
from pathlib import Path

from setuptools import Command, setup

try:
    import pip._internal.req as req
except ImportError:
    import pip.req as req

try:
    import pypandoc
except ImportError:
    pypandoc = None


if sys.version_info < (3, 6):
    sys.stderr.write('Python 3.6+ is required for installation.\n')
    sys.tracebacklimit = 0
    sys.exit(1)


def main():
    """Run setup"""
    metadata = get_package_metadata()

    # Run setup
    setup(**metadata)


class CleanCommand(Command, object):
    description = 'Remove build artifacts and *.pyc'
    user_options = list()

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        repo_path = str(Path(__file__).parent)
        removables = [
            'build', '_build', 'dist', 'wheelhouse',
            '*.egg-info', '*.egg', '.eggs'
            ]
        for removable in removables:
            for path in os.scandir(repo_path):
                if Path(path).match(removable):
                    if path.is_dir():
                        shutil.rmtree(path)
                    elif path.is_file():
                        os.remove(path)

        for root, folders, files in os.walk(repo_path):
            for folder in folders:
                if folder == '__pycache__':
                    fpath = os.path.join(root, folder)
                    shutil.rmtree(fpath)
            for filename in files:
                fpath = os.path.join(root, filename)
                if Path(filename).match('*.py[co]'):
                    os.remove(fpath)


# ----------------------------------------------------------------------
# Support
# ----------------------------------------------------------------------
files_in_tree = set()
folders_in_tree = set()


def find_data_files(repo_path=None):
    """Captures files that are project specific"""
    repo_path = repo_path or Path(__file__).parent

    include = ['LICENSE', 'requirements', 'requirements*.txt', 'entrypoints.txt']
    found = [
        str(path)
        for path in scan_tree(repo_path)
        if (any(path.match(included) for included in include)
            or any(parent.match(included) for parent in path.parents for included in include))
        ]
    return found


def find_packages(repo_path=None):
    """Finds packages; replacement for setuptools.find_packages which
    doesn't support PEP 420.

    Lengthy discussion with no resolution:
        https://github.com/pypa/setuptools/issues/97

    Assumptions:
      * Packages are folders
      * Modules are files
      * Namespaces are folders without modules
      * setup.py is for a single package
      * the folder which contains setup.py is _not_ a package itself
      * each package falls under a single tree (e.g. requests, requests.api, etc.)
      * the package is complex enough to have multiple sub folders/modules

    Args:
        repo_path (str): path to check [default: path of setup.py]

    Returns:
        list(list, list, list): Returns packages, modules and namespaces

    """
    repo_path = repo_path or Path(__file__).parent
    packages = set()
    modules = set()
    namespaces = set()
    python_artifacts = ['__pycache__']
    repo_artifacts = ['.*']
    install_artifacts = ['*.egg-info']
    build_artifacts = ['dist', 'build']
    test_files = ['tests', 'test', '.tox']
    artifacts = install_artifacts + build_artifacts + repo_artifacts + python_artifacts + test_files
    for path in scan_tree(str(repo_path)):
        # Only include paths with .py
        if not path.match('*.py'):
            continue

        # check folders
        if any(p.match(artifact) for artifact in artifacts for p in path.parents):
            continue

        # There should be a valid package at this point
        package_name = str(path.parent).replace(os.path.sep, '.')
        module_name = str(path).replace(os.path.sep, '.').replace(path.suffix, '')
        if package_name == '.':
            module_name = '.' + module_name
            if module_name == '.setup':
                continue
            modules.add(module_name)
            continue

        modules.add(module_name)
        packages.add(package_name)

    # find namespaces
    for module_name in modules:
        module_path = Path(module_name.replace('.', os.path.sep))
        package_name = str(module_path.parent).replace(os.path.sep, '.')
        if package_name in packages:
            continue
        namespaces.add(package_name)

    return sorted(packages), sorted(modules), sorted(namespaces)


def get_entrypoints(path=None):
    """
    """
    entry_points = {}
    path = path or 'entrypoints.txt'
    if not os.path.exists(path):
        return entry_points

    with open(path) as stream:
        for line in stream:
            if not line.strip():
                continue
            if line.strip().startswith('#'):
                continue
            entry_points.setdefault('console_scripts', []).append(line.strip())
    return entry_points


def get_license(top_path=None):
    """Reads license file and returns"""
    repo_path = top_path or os.path.realpath(os.path.dirname(__file__))
    files = {f.lower(): f for f in os.listdir(repo_path)}
    permutations = itertools.product(['license'], ['', '.txt'])
    files = [os.path.join(repo_path, f) for l, f in files.items() if l in permutations]
    license = ''
    for filepath in files:
        if os.path.exists(filepath):
            with open(filepath, 'r') as stream:
                license = stream.read()
                break
    return license


def get_package_metadata(top_path=None):
    """Find the __metadata__.py file and read it"""
    repo_path = top_path or os.path.realpath(os.path.dirname(__file__))
    metadata = {}
    prefixes = ('.', '_')
    for root, folders, files in os.walk(repo_path):
        rel = root.replace(repo_path, '').lstrip(os.path.sep)
        folders[:] = [
            folder for folder in folders
            if not any(folder.startswith(prefix) for prefix in prefixes)
            ]
        for filename in files:
            filepath = Path(os.path.join(rel, filename))
            if filepath.name == '__metadata__.py':
                exec(filepath.open().read(), globals(), metadata)
                metadata = metadata.get('package_metadata') or metadata
                break
        if metadata:
            break

    requirements, dependency_links = get_package_requirements(top_path=top_path)
    packages, modules, namespaces = find_packages()
    # Package Properties
    metadata.setdefault('long_description', metadata.get('doc') or get_readme())
    metadata.setdefault('packages', packages)
    metadata.setdefault('include_package_data', True)

    # Requirements
    metadata.setdefault('setup_requires', requirements.get('setup') or [])
    metadata.setdefault('install_requires', requirements.get('install') or [])
    metadata.setdefault('tests_require', requirements.get('tests') or requirements.get('test') or [])
    metadata.setdefault('extras_require', requirements.get('extras') or [])
    metadata.setdefault('dependency_links', dependency_links)

    # CLI
    entry_points = get_entrypoints() or {}
    metadata.setdefault('entry_points', entry_points)

    # Packaging
    metadata.setdefault('platforms', ['any'])
    metadata.setdefault('zip_safe', False)

    year = datetime.datetime.now().year
    license = get_license() or 'Copyright {year} - all rights reserved'.format(year=year)
    metadata.setdefault('license', license)

    # Extra data
    metadata.setdefault('data_files', [('', find_data_files())])

    # Add setuptools commands
    metadata.setdefault('cmdclass', get_setup_commands())
    return metadata


def get_package_requirements(top_path=None):
    """Find all of the requirements*.txt files and parse them"""
    repo_path = top_path or os.path.realpath(os.path.dirname(__file__))
    requirements = {'extras': {}}
    dependency_links = set()
    # match on:
    #    requirements.txt
    #    requirements-<name>.txt
    #    requirements_<name>.txt
    #    requirements/<name>.txt
    options = '_-/'
    include_globs = ['requirements*.txt', 'requirements/*.txt']
    paths = [
        relpath
        for relpath in scan_tree(repo_path, include=include_globs)
        ]
    for path in paths:
        if 'requirements' in map(str, path.parents):
            name = path.name.replace(path.suffix, '')
        elif 'requirements' in path.name:
            name = path.name.replace('requirements', '').lstrip(options)
        else:
            raise Exception('Could not find requirements using ' + path)
        reqs_, deps = parse_requirements(str(path.absolute()))
        dependency_links.update(deps)
        if name in ['install', '']:
            requirements['install'] = reqs_
        elif name in ['test', 'tests']:
            requirements['tests'] = reqs_
            requirements['extras']['tests'] = reqs_
        elif name in ['setup']:
            requirements['setup'] = reqs_
        else:
            requirements['extras'][name] = reqs_

    all_reqs = set()
    dev_reqs = set()
    for name, req_list in requirements.items():
        if name in ['install']:
            all_reqs.update(req_list)
        elif name in ['extras']:
            for subname, reqs in req_list.items():
                all_reqs.update(reqs)
                dev_reqs.update(reqs)
        else:
            all_reqs.update(req_list)
            dev_reqs.update(req_list)

    requirements['extras']['dev'] = list(sorted(dev_reqs))
    requirements['extras']['all'] = list(sorted(all_reqs))
    return requirements, list(sorted(dependency_links))


def get_readme(top_path=None):
    """Read the readme for the repo"""
    path = top_path or os.path.realpath(os.path.dirname(__file__))
    found = {f.name.lower(): f.name for f in os.scandir(path)}
    permutations = [a + b for a, b in itertools.product(['readme'], ['.md', '.rst', '.txt'])]
    files = [os.path.join(path, f) for l, f in found.items() if l in permutations]
    readme = ''
    for filepath in files:
        if pypandoc and filepath.endswith('.md'):
            readme = pypandoc.convert(filepath, 'rst')
            break
        else:
            with open(filepath, 'r') as stream:
                readme = stream.read()
                break
    return readme


def get_setup_commands():
    """Returns setup command class list"""
    commands = {
        'clean': CleanCommand,
        }
    return commands


def parse_requirements(path):
    template = '{name}{spec}'
    requirements = set()
    dependency_links = set()
    for requirement in req.parse_requirements(path, session="somesession"):
        if requirement.markers is not None and not requirement.markers.evaluate():
            continue

        name = requirement.name
        spec = str(requirement.req.specifier) if len(str(requirement.req.specifier)) else ''
        req_ = template.format(name=name, spec=spec)
        if req_:
            requirements.add(req_)

        link = str(requirement.link) if requirement.link else ''
        if link:
            dependency_links.add(link)

        # TODO: What do we do with these?
        if requirement.options:
            pass

    return list(sorted(requirements)), list(sorted(dependency_links))


def scan_tree(top_path=None, include=None):
    """Finds files in tree

    * Order is random
    * Folders which start with . and _ are excluded unless excluded is used (e.g. [])
    * This list is memoized

    Args:
        top_path (str): top of folder to search
        include (list, optional): filters on include if present

    Yields:
        str: paths as found
    """
    repo_path = Path(__file__)
    if not files_in_tree:
        for root, folders, files in os.walk(str(top_path or repo_path)):
            rel = Path(root.replace(str(top_path or repo_path), '').lstrip('/'))
            # Control traversal
            folders[:] = [f for f in folders if f not in ['.git']]
            folders_in_tree.update(folders)
            # Yield files
            for filename in files:
                relpath = rel.joinpath(filename)
                if relpath not in files_in_tree:
                    files_in_tree.add(relpath)
                    if include is not None:
                        if any(relpath.match(inc) for inc in include):
                            yield relpath
                    else:
                        yield relpath
    else:
        for relpath in files_in_tree:
            if include:
                if any(relpath.match(inc) for inc in include):
                    yield relpath
            else:
                yield relpath


if __name__ == '__main__':
    main()



import sys
import textwrap


def echo(*messages, verbose=None, indent=True, end=None, flush=None, file=None):
    """Echo *message*.

    indentation can be finely controlled, but by default, indentation
    is related to the verbose setting set by the command-line interface
    and the verbose setting passed in as a parameter.  The difference
    of the two identifies how much indentation is created.

    Args:
        messages (Tuple[Any]): messages to echo
        verbose (bool|int, optional): verbosity level
        indent (int): number of spaces to indent
        end (str): text to print at end of message, defaults to newline
        flush (bool): flush output after write
        file (typing.IO): file to echo message to

    Returns:
        Tuple[str]: beautified messages

    """
    verbose = True if verbose is None else max(int(verbose or 0), 0)
    indent = 0 if indent in [None, False, True] else max(int(indent or 0), 0)
    file = file or sys.stdout
    messages = list(messages)
    if verbose:
        for index, message in enumerate(messages):
            message_end = end if index == len(messages) - 1 else ' '
            message = '' if message is None else message
            end = '\n' if end is None else end
            if end == '\n' and indent:
                message = textwrap.indent(message, ' ' * indent)
            message = message.rstrip('\n') if end == '\n' else message
            print(message, flush=flush, end=message_end, file=file)
    return tuple(messages)
